@@ -1,25 +1,27 @@
 import argparse
 import copy
+import datetime
 import json
-import os
 import sys
 import time
-import datetime
-from shutil import which
-from typing import Union
+from typing import Optional
 
-from termcolor import colored
 from docker.errors import ImageNotFound, APIError
+from termcolor import colored
 
+from cpk import CPKProject, cpkconfig
+from cpk.utils.docker import transfer_image
+from cpk.utils.misc import human_time, configure_binfmt
+from .endpoint import CLIEndpointInfoCommand
 from .info import CLIInfoCommand
 from .. import AbstractCLICommand
 from ..logger import cpklogger
-from cpk import CPKProject
-from cpk.utils.docker import get_client, get_endpoint_ncpus, DOCKER_INFO, \
-    get_endpoint_architecture, pull_image
-from cpk.utils.misc import sanitize_hostname, human_size, human_time, configure_binfmt
+from ...constants import ARCH_TO_DOCKER_PLATFORM
 from ...exceptions import CPKProjectBuildException
+from ...types import Machine, Arguments
+from ...utils.cli import check_git_status
 from ...utils.image_analyzer import EXTRA_INFO_SEPARATOR, ImageAnalyzer, SEPARATORS_LENGTH
+from ...utils.machine import get_machine
 
 
 class CLIBuildCommand(AbstractCLICommand):
@@ -27,7 +29,8 @@ class CLIBuildCommand(AbstractCLICommand):
     KEY = 'build'
 
     @staticmethod
-    def parser(parent: Union[None, argparse.ArgumentParser] = None) -> argparse.ArgumentParser:
+    def parser(parent: Optional[argparse.ArgumentParser] = None,
+               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
         parser = argparse.ArgumentParser(parents=[parent])
         parser.add_argument(
             "--pull",
@@ -87,12 +90,14 @@ class CLIBuildCommand(AbstractCLICommand):
             action="store_true",
             help="Stamp image with the build time"
         )
+        # TODO: to be implemented
         parser.add_argument(
             "-D",
             "--destination",
             default=None,
-            help="Docker socket or hostname where to deliver the image"
+            help="CPK machine or endpoint hostname where to deliver the image once built"
         )
+        # TODO: to be implemented
         parser.add_argument(
             "--docs",
             default=False,
@@ -114,47 +119,32 @@ class CLIBuildCommand(AbstractCLICommand):
         return parser
 
     @staticmethod
-    def execute(parsed: argparse.Namespace) -> bool:
+    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
         stime = time.time()
-        parsed.workdir = os.path.abspath(parsed.workdir)
 
         # get project
         project = CPKProject(parsed.workdir, parsed=parsed)
 
         # show info about project
-        CLIInfoCommand.execute(parsed)
+        CLIInfoCommand.execute(None, parsed)
 
-        # check if the git HEAD is detached
-        if project.is_detached():
-            cpklogger.error(
-                "The repository HEAD is detached. Create a branch or check one out "
-                "before continuing. Aborting."
-            )
+        # check git workspace status
+        proceed = check_git_status(project, parsed)
+        if not proceed:
             return False
 
-        # check if the index is clean
-        if project.is_dirty():
-            cpklogger.warning("Your index is not clean (some files are not committed).")
-            cpklogger.warning("If you know what you are doing, use --force (-f) to force.")
-            if not parsed.force:
-                return False
-            cpklogger.warning("Forced!")
-
-        # sanitize hostname
-        if parsed.machine is not None:
-            parsed.machine = sanitize_hostname(parsed.machine)
-
-        # create docker client
-        docker = get_client(parsed.machine)
-
         # get info about docker endpoint
-        cpklogger.info("Retrieving info about Docker endpoint...")
-        epoint = docker.info()
-        if "ServerErrors" in epoint:
-            cpklogger.error("\n".join(epoint["ServerErrors"]))
-            return False
-        epoint["MemTotal"] = human_size(epoint["MemTotal"])
-        cpklogger.print(DOCKER_INFO.format(**epoint))
+        CLIEndpointInfoCommand.execute(machine, parsed)
+
+        # pick right value of `arch` given endpoint
+        machine_arch = machine.get_architecture()
+        if parsed.arch is None:
+            cpklogger.info("Parameter `arch` not given, will resolve it from the endpoint.")
+            parsed.arch = machine_arch
+            cpklogger.info(f"Parameter `arch` automatically set to `{parsed.arch}`.")
+
+        # create docker client
+        docker = machine.get_client()
 
         # define build-args
         buildargs = {"buildargs": {}, "labels": {}}
@@ -170,22 +160,17 @@ class CLIBuildCommand(AbstractCLICommand):
         buildargs["labels"].update(project.build_labels())
         # - build-arg NCPUS
         buildargs['buildargs']['NCPUS'] = \
-            str(get_endpoint_ncpus(docker)) if parsed.ncpus is None else str(parsed.ncpus)
-
-        # pick the right architecture if not set
-        if parsed.arch is None:
-            parsed.arch = get_endpoint_architecture(parsed.machine)
-            cpklogger.info(f"Target architecture automatically set to {parsed.arch}.")
+            str(parsed.ncpus) if parsed.ncpus else str(machine.get_ncpus())
 
         # create defaults
         image = project.image(parsed.arch)
 
         # print info about multiarch
-        msg = "Building an image for {} on {}.".format(parsed.arch, epoint["Architecture"])
-        cpklogger.info(msg)
+        cpklogger.info("Building an image for {} on {}.".format(parsed.arch, machine_arch))
         # - register bin_fmt in the target machine (if needed)
         if not parsed.no_multiarch:
-            configure_binfmt(parsed.arch, docker, cpklogger)
+            configure_binfmt(machine_arch, parsed.arch, docker, cpklogger)
+        platform = ARCH_TO_DOCKER_PLATFORM.get(parsed.arch, None)
 
         # architecture target
         buildargs["buildargs"]["ARCH"] = parsed.arch
@@ -211,7 +196,7 @@ class CLIBuildCommand(AbstractCLICommand):
                 # try to pull the same image so Docker can use it as cache source
                 cpklogger.info(f'Pulling image "{image}" to use as cache...')
                 try:
-                    pull_image(image, endpoint=docker)
+                    machine.pull_image(image)
                     is_present = True
                 except KeyboardInterrupt:
                     cpklogger.info("Aborting.")
@@ -242,6 +227,7 @@ class CLIBuildCommand(AbstractCLICommand):
                 local_sha = project.version.sha
                 # get remote image metadata
                 try:
+                    # TODO: parsed.machine is now NONE
                     labels = project.image_labels(parsed.machine, parsed.arch)
                     time_label = project.label("time")
                     sha_label = project.label("code.sha")
@@ -268,6 +254,7 @@ class CLIBuildCommand(AbstractCLICommand):
                 "pull": parsed.pull,
                 "nocache": parsed.no_cache,
                 "tag": image,
+                "platform": platform
             }
         )
         cpklogger.debug("Build arguments:\n%s\n" % json.dumps(buildargs, sort_keys=True, indent=4))
@@ -301,12 +288,12 @@ class CLIBuildCommand(AbstractCLICommand):
             buildlog.append(msg)
             cpklogger.print(msg)
 
-        # build code docs
-        if parsed.docs:
-            docs_args = ["--quiet"] * int(not parsed.verbose)
-            # build docs
-            cpklogger.info("Building documentation...")
-            # TODO: fix this
+        # TODO: fix this
+        # # build code docs
+        # if parsed.docs:
+        #     docs_args = ["--quiet"] * int(not parsed.verbose)
+        #     # build docs
+        #     cpklogger.info("Building documentation...")
 
         # get image history
         historylog = [(layer["Id"], layer["Size"]) for layer in dimage.history()]
@@ -338,10 +325,10 @@ class CLIBuildCommand(AbstractCLICommand):
             buildlog, historylog, extra_info=extra_info
         )
         # pull image (if the destination is different from the builder machine)
-        if parsed.destination and parsed.machine != parsed.destination:
-            _transfer_image(
-                origin=parsed.machine,
-                destination=parsed.destination,
+        if parsed.destination and machine.base_url != parsed.destination:
+            transfer_image(
+                origin=machine,
+                destination=get_machine(parsed.destination, cpkconfig.machines),
                 image=image,
                 image_size=final_image_size,
             )
@@ -349,7 +336,7 @@ class CLIBuildCommand(AbstractCLICommand):
         if parsed.push:
             # call command `push`
             from .push import CLIPushCommand
-            CLIPushCommand.execute(copy.deepcopy(parsed))
+            CLIPushCommand.execute(machine, copy.deepcopy(parsed))
 
         # perform remove (if needed)
         if parsed.rm:
@@ -357,23 +344,9 @@ class CLIBuildCommand(AbstractCLICommand):
             # noinspection PyBroadException
             try:
                 # call command `clean`
-                CLICleanCommand.execute(copy.deepcopy(parsed))
+                CLICleanCommand.execute(machine, copy.deepcopy(parsed))
             except BaseException:
-                cpklogger.warn(
-                    "We had some issues cleaning up the image on '{:s}'".format(parsed.machine)
-                    + ". Just a heads up!"
-                )
-
-
-def _transfer_image(origin, destination, image, image_size):
-    monitor_info = "" if which("pv") else " (install `pv` to see the progress)"
-    cpklogger.info(f'Transferring image "{image}": [{origin}] -> [{destination}]{monitor_info}...')
-    data_source = ["docker", "-H=%s" % origin, "save", image]
-    data_destination = ["docker", "-H=%s" % destination, "load"]
-    progress_monitor = ["|", "pv", "-cN", "image", "-s", image_size] if which("pv") else []
-    cmd = data_source + progress_monitor + data_destination
-    # TODO: re-enable this
-    # start_command_in_subprocess(cmd, nostdout=True)
+                cpklogger.warn("We had some issues cleaning up the image. Just a heads up!")
 
 
 def _build_line(line):
@@ -391,5 +364,3 @@ def _build_line(line):
         line += "\n"
     # ---
     return line
-
-
